# source: https://pubs.acs.org/doi/10.1021/acs.macromol.9b00262

from __future__ import annotations

import warnings
from collections import Counter
from typing import Iterable, Tuple

import numpy as np
from pylimer_tools_cpp import Molecule, MoleculeType, Universe


def predictShearModulus(networks: Iterable[Universe], T: float = 1, k_B: float = 1, foreignAtomType=None, totalMass=1) -> float:
    """
    Predict the shear modulus using ANT Analysis.

    Source:
      - https://pubs.acs.org/doi/10.1021/acs.macromol.9b00262

    Arguments:
      - network: the polymer system to predict the shear modulus for
      - T: the temperature in your unit system
      - k_b: Boltzmann's constant in your unit system
      - foreignAtomType: the type of atoms to ignore (junctions, crosslinkers)
      - totalMass: the :math:`M` in the respective formula

    Returns:
      - shear modulus (float): the estimated shear modulus. Unit: [pressure]
    """
    Gamma = calculateTopologicalFactor(networks, foreignAtomType, totalMass)
    nu = 0
    for network in networks:
        nu += len(network.getMolecules(foreignAtomType)) / \
            (network.getVolume()) / len(networks)
    return Gamma*nu*k_B*T


def calculateCycleRank(networks: Iterable[Universe] = None, nu: int = None, mu: int = None, absTol: float = 1, relTol: float = 1, junctionType=None) -> float:
    """
    Compute the cycle rank (:math:`\\chi`).
    Assumes the precursor-chains to be bifunctional.

    Arguments:
      - network: the network to calculate the cycle rank for
      - nu: number of elastically effective (active) strands per unit volume
      - mu: number density of the elastically effective crosslink
      - absTol (float): the absolute tolerance to categorize a chain as active (min. end-to-end distance) (None to use only relTol)
      - relTol (float): the relative tolerance to categorize a chain as active (0: all, 1: none (use only absTol))
      - junctionType: the atom type of the crosslinkers/junctions

    No need to provide all the parameters — either/or:
    - nu & mu
    - network, absTol, relTol, junctionType

    Returns:
      - cycleRank: the cycle rank ($\\xi = \\nu_{eff} - \\mu_{eff}$). Unit: [1/Volume]
    """
    if (nu is None):
        if (junctionType is None or networks is None):
            raise ValueError(
                "Argument missing: When not specifiying nu, network and junctionType need to be specified")
        nu = calculateEffectiveNrDensityOfNetwork(
            networks, absTol, relTol, junctionType)
    if (mu is None):
        if (junctionType is None or networks is None):
            raise ValueError(
                "Argument missing: When not specifiying mu, network and junctionType need to be specified")
        mu = calculateEffectiveNrDensityOfJunctions(
            networks, absTol, relTol, junctionType)

    return nu - mu


def calculateEffectiveNrDensityOfNetwork(networks: Iterable[Universe], absTol: float = 1, relTol: float = 1, junctionType=None) -> float:
    """
    Compute the effective number density :math:`\\nu_{eff}` of a network.
    Assumes the precursor-chains to be bifunctional.

    :math:`\\nu_{eff}` is the number of elastically effective (active) strands per unit volume,
    which are defined as the ones that can store elastic energy
    upon network deformation, resp. the effective number density of network strands

    Source:
      - https://pubs.acs.org/doi/10.1021/acs.macromol.9b00262

    Arguments:
      - network (pylimer_tools.entities.Universe): the network to compute :math:`\\nu_{eff}` for
      - absTol (float): the absolute tolerance to categorize a chain as active (min. end-to-end distance) (None to use only relTol)
      - relTol (float): the relative tolerance to categorize a chain as active (0: all, 1: none (use only absTol))
      - junctionType: the atom type of the crosslinkers/junctions

    Returns:
      - :math:`\\nu_{eff}` (float): the effective number density of network strands. Unit: [1/Volume]
    """
    if (len(networks) == 0):
        return None

    # get the mean end to end distances
    R_taus = computeMeanEndToEndDistances(networks, junctionType)
    if (len(R_taus) < 1):
        return 0.0
    R_taus = np.array(list(R_taus.values()))
    R_tau_max = np.max(R_taus)

    # process additional input parameters
    if (absTol is None):
        absTol = R_tau_max

    # count how many effective strands there are
    numEffective = np.array([R_tau > absTol or R_tau > relTol*R_tau_max
                             for R_tau in R_taus]).sum()
    meanVolume = calculateMeanUniverseVolume(networks)

    return numEffective / meanVolume


def calculateMeanUniverseVolume(networks: Iterable[Universe], acceptDifferentSizes: bool = False) -> float:
    """
    Compute the mean volume of a list of universes.

    Arguments:
      - networks: a list of universes
      - acceptDifferentSizes: toggle whether to throw an error when the Universe have different nr. of atoms

    Returns:
      - meanVolume (float): the mean volume of the universes
    """
    # compute the mean volume of the universes
    meanVolume = 0
    divisor = 1/len(networks)
    networkSize = networks[0].getNrOfAtoms()
    for network in networks:
        if (not acceptDifferentSizes and network.getNrOfAtoms() != networkSize):
            raise NotImplementedError(
                "Currently, only sequences of networks with the same size are supported (got one with {} instead of {})".format(network.getNrOfAtoms(), networkSize))
        meanVolume += network.getVolume()*divisor
    return meanVolume


def calculateEffectiveNrDensityOfJunctions(networks: Iterable[Universe], absTol: float = 0, relTol: float = 1, junctionType=None, minNumEffectiveStrands=2) -> float:
    """
    Compute the number density of the elastically effective crosslinks,
    defined as the ones that connect at least two elastically effective strands.
    Assumes the precursor-chains to be bifunctional.

    Source:
      - https://pubs.acs.org/doi/10.1021/acs.macromol.9b00262

    Arguments:
      - network (pylimer_tools.entities.Universe): the network to compute :math:`\\nu_{eff}` for
      - absTol (float): the absolute tolerance to categorize a chain as active (min. end-to-end distance) (None to use only relTol)
      - relTol (float): the relative tolerance to categorize a chain as active (0: all, 1: none (use only absTol))
      - junctionType: the atom type of the crosslinkers/junctions
      - minNumEffectiveStrands (int): the number of elastically effective strands to qualify a junction as such

    Returns:
      - :math:`\\mu_{eff}` (float): the effective number density of junctions. Unit: [1/Volume]
    """
    if (len(networks) < 1):
        return None
    if (junctionType is None):
        return 0.0

    meanVolume = calculateMeanUniverseVolume(networks)

    if (minNumEffectiveStrands == 0):
        return len(networks[0].getAtomsWithType(junctionType))/meanVolume

    # get the mean end to end distances
    R_taus = computeMeanEndToEndDistances(networks, junctionType)
    if (len(R_taus) < 1):
        return 0.0
    R_tau_max = max(R_taus.values())

    # process additional input parameters
    if (absTol is None):
        absTol = R_tau_max

    # count how many active connections each junction has
    junctionActivity = {}
    for key in R_taus:
        crosslinkerNames = key.split("+")
        assert(len(crosslinkerNames) == 3)
        isActive = R_taus[key] > absTol or R_taus[key] > relTol*R_tau_max
        if (not(isActive)):
            continue
        relevantNames = [crosslinkerNames[0], crosslinkerNames[1]]
        for crosslinkerName in relevantNames:
            if (crosslinkerName not in junctionActivity):
                junctionActivity[crosslinkerName] = 0
            junctionActivity[crosslinkerName] += 1

    effectiveJunctions = np.array(
        [junctionActivity[key] >= minNumEffectiveStrands for key in junctionActivity])
    numEffectiveJunctions = effectiveJunctions.sum()
    return numEffectiveJunctions/meanVolume


def calculateWeightFractionOfBackbone(network: Universe, crosslinkerType):
    """
    Compute the weight fraction of network backbone in infinite network

    Arguments:
      - network: the network to compute the weight fraction for
      - crosslinkerType: the atom type to use to split the molecules

    Returns:
      - weightFraction (float): 1 - weightDangling/weightTotal,
    """
    weightFraction, _ = calculateWeightFractionOfDanglingChains(
        network, crosslinkerType)
    return 1.0 - weightFraction


def calculateWeightFractionOfDanglingChains(network: Universe, crosslinkerType) -> Tuple[float, float]:
    """
    Compute the weight fraction of dangling strands in infinite network

    Arguments:
      - network: the network to compute the weight fraction for
      - crosslinkerType: the atom type to use to split the molecules

    Returns:
      - weightFraction: weightDangling/weightTotal,
      - numFraction: numDangling/numTotal
    """
    if (network.getNrOfAtoms() < 1):
        return 0.0, 0.0

    weights = network.getMasses()

    def getWeightOfGraph(graph):
        counts = Counter(graph.getAtomTypes())
        weightTotal = 0
        for key in counts:
            weightTotal += weights[key]*counts[key]
        return weightTotal

    allChains = network.getChainsWithCrosslinker(crosslinkerType)
    numTotal = network.getNrOfAtoms()
    weightTotal = getWeightOfGraph(network)

    numDangling = 0
    weightDangling = 0
    for chain in allChains:
        if (chain.getType() == MoleculeType.DANGLING_CHAIN):
            numDangling += chain.getNrOfAtoms()
            weightDangling += getWeightOfGraph(chain)

    if (weightTotal == 0):
        # warnings.warn("Total weight of network is = 0.")
        return 0.0, numDangling/numTotal

    return weightDangling/weightTotal, numDangling/numTotal


def computeMeanEndToEndDistances(networks: Iterable[Universe], crosslinkerType) -> dict:
    """
    Compute the mean end to end distance between each pair of (indirectly) connected crosslinker

    Arguments:
      - networks: the different configurations of the polymer network to do the computation for
      - crosslinkerType: the atom type to compute the in-between vectors for

    Returns:
      - endToEndDistances (dict): a dictionary with key: "{atom1.name}+{atom2.name}"
          and value: the norm of the mean difference vector
    """
    R_tau_vectors = computeMeanEndToEndVectors(networks, crosslinkerType)
    if (len(R_tau_vectors) < 1):
        return {}

    R_tau_vectors_array = np.array(list(R_tau_vectors.values()))
    R_taus = np.linalg.norm(R_tau_vectors_array, axis=1)

    return dict(zip(R_tau_vectors.keys(), R_taus))


def computeMeanEndToEndVectors(networks: Iterable[Universe], crosslinkerType) -> dict:
    """
    Compute the mean end to end vectors between each pair of (indirectly) connected crosslinker

    Arguments:
      - networks: the different configurations of the polymer network to do the computation for
      - crosslinkerType: the atom type to compute the in-between vectors for

    Returns:
      - endToEndVectors (dict): a dictionary with key: "{atom1.name}+{atom2.name}"
          and value: their mean distance difference vector
    """
    if (len(networks) == 0):
        return {}
    endToEndVectors = {}
    divider = 1/len(networks)
    for network in networks:
        currentEndToEndVectors = computeEndToEndVectors(
            network, crosslinkerType)
        # the mean calculation in this for loop
        # trades some memory for performance
        # there are still many performance and memory
        # improvements possible
        # (e.g. computing connectivity only once, storing it only once, ....)
        for key in currentEndToEndVectors:
            if (key not in endToEndVectors):
                endToEndVectors[key] = [0, 0, 0]
            for i in range(3):
                endToEndVectors[key][i] += currentEndToEndVectors[key][i]*divider
    return endToEndVectors


def computeEndToEndVectors(network: Universe, crosslinkerType) -> dict:
    """
    Compute the end to end vectors between each pair of (indirectly) connected crosslinker

    Arguments:
      - network: the polymer network to do the computation for
      - crosslinkerType: the atom type to compute the in-between vectors for

    Returns:
      - endToEndVectors (dict): a dictionary with key: "{atom1.name}+{atom2.name}"
          and value: their difference vector
    """
    # while we could do the decomposition again with explicit removal of irrelevant strand atoms,
    # this should not be any more expensive
    endToEndVectors = {}
    molecules = network.getChainsWithCrosslinker(crosslinkerType)
    for molecule in molecules:
        crosslinkers = molecule.getAtomsWithType(crosslinkerType)
        if (len(crosslinkers) != 2 or
            molecule.getType() == MoleculeType.PRIMARY_LOOP or
                molecule.getType() == MoleculeType.DANGLING_CHAIN):
            # dangling, free chains and loops are irrelevant for our purposes
            continue
        # igraph.VertexSeq is not sortable -> use a list
        crosslinkers = [crosslinkers[0], crosslinkers[1]]
        # sort crosslinkers by name as a way to keep the vector directions consistent between timesteps
        crosslinkers.sort(key=lambda a: a.getId())
        #
        key = _getKeyForMolecule(molecule, crosslinkers)
        endToEndVectors[key] = crosslinkers[0].computeVectorTo(
            crosslinkers[1], network.getBox())

    return endToEndVectors


def computeCrosslinkerConversion(network: Universe, junctionType, f: int = None) -> float:
    """
    Compute the extent of reaction of the crosslinkers
    (actual functionality divided by target functionality)

    Arguments:
      - network: the polymer network to do the computation for
      - junctionType: the type of the junctions/crosslinkers to select them in the network
      - f: the functionality of the crosslinkers

    Returns:
      - r (float): the (mean) crosslinker conversion
    """
    if (f is None):
      f = network.determineFunctionalityPerType()[junctionType]
    return calculateEffectiveCrosslinkerFunctionality(network, junctionType) / f


def calculateEffectiveCrosslinkerFunctionality(network: Universe, junctionType) -> float:
    """
    Compute the mean crosslinker functionality

    Arguments:
      - network: the polymer network to do the computation for
      - junctionType: the type of the junctions/crosslinkers to select them in the network

    Returns:
      - f (float): the (mean) effective crosslinker functionality
    """
    junctionDegrees = calculateEffectiveCrosslinkerFunctionalities(
        network, junctionType)
    return np.mean(junctionDegrees)


def calculateEffectiveCrosslinkerFunctionalities(network: Universe, junctionType) -> list[int]:
    """
    Compute the functionality of every crosslinker in the network

    Arguments:
      - network: the polymer network to do the computation for
      - junctionType: the type of the junctions/crosslinkers to select them in the network

    Returns:
      - junctionDegrees (list[int]): the functionality of every crosslinker
    """
    if (network.getNrOfAtoms() == 0):
        return []
    junctions = network.getAtomsWithType(junctionType)
    junctionIds = [v.getId() for v in junctions]
    junctionDegrees = [network.getNrOfBondsOfAtom(id) for id in junctionIds]
    return junctionDegrees


def calculateTopologicalFactor(networks: Iterable[Universe], foreignAtomType=None, totalMass=1, b=None) -> float:
    """
    Compute the topological factor of a polymer network.

    Assumptions: 
      - the precursor-chains to be bifunctional
      - all Universes to have the same structure (with possibly differing positions)
      - crosslinkers do not count to the nr. of monomers in a strand

    Source:
      - eq. 16 in https://pubs.acs.org/doi/10.1021/acs.macromol.9b00262

    Arguments:
      - network: the network to compute the topological factor for
      - foreignAtomType: the type of atoms to ignore
      - totalMass: the :math:`M` in the respective formula
      - b: the mean bond length. 
          If `None`, it will be computed for each molecule in the first Universe (Network).

    Returns:
      - the topological factor :math:`\\Gamma`
    """
    R_taus = computeMeanEndToEndDistances(networks, foreignAtomType)

    # find the topological factor
    GammaSum = 0
    network = networks[0]  # this is where the second assumption is made
    chainsToProcess = network.getChainsWithCrosslinker(foreignAtomType)
    for molecule in chainsToProcess:
        crosslinkers = molecule.getAtomsWithType(foreignAtomType)
        if (len(crosslinkers) != 2 or
            molecule.getType() == MoleculeType.PRIMARY_LOOP or
                molecule.getType() == MoleculeType.DANGLING_CHAIN):
            # dangling, free chains and loops are irrelevant for our purposes
            continue
        if (b is None):
            b = np.mean(molecule.computeBondLengths())
        crosslinkers = [crosslinkers[0], crosslinkers[1]]
        # sort crosslinkers by name as a way to keep the vector directions consistent between timesteps
        crosslinkers.sort(key=lambda a: a.getId())
        key = _getKeyForMolecule(molecule, crosslinkers)
        GammaSum += R_taus[key]*R_taus[key] / \
            ((molecule.getLength()-2) * b *
             b)  # -2: remove crosslinkers again (assumption 3)

    return GammaSum / totalMass


def _getKeyForMolecule(molecule, crosslinkers):
    """
    Get a key to identify a molecule.
    The crosslinkers (ends of the molecule) are the first two components of the key, 
    whereas the names of all other atoms in the chain are used too to distinguish 
    e.g. two secondary loops
    """
    names = [a.getId() for a in molecule.getAtoms()]
    names.sort()
    return "{}+{}+{}".format(crosslinkers[0].getId(), crosslinkers[1].getId(), "-".join(str(n) for n in names))




import math
import warnings
from collections import Counter

import numpy as np
from pylimer_tools_cpp import Universe
from scipy import optimize


def predictShearModulus(network: Universe, junctionType, strandLength: int = None, functionalityPerType=None, T: float = 1, k_B: float = 1):
    """
    Predict the shear modulus using MMT Analysis.

    Source:
      - https://pubs.acs.org/doi/10.1021/acs.macromol.9b00262

    Arguments:
      - network: the polymer system to predict the shear modulus for
      - junctionType: the type of atoms making up the junctions/crosslinkers
      - junctionType: the type of the junctions/crosslinkers to select them in the network
      - strandLength: the length of the network strands (in nr. of beads). See: #computeStoichiometricInbalance
      - T: the temperature in your unit system
      - k_b: Boltzmann's constant in your unit system
      - totalMass: the :math:`M` in the respective formula

    Returns:
      - G: the predicted shear modulus, or `None` if the universe is empty.

    ToDo:
      - Support more than one crosslinker type (as is supported by original formula)
    """
    if (network.getNrOfAtoms() == 0):
        return None
    nu = len(network.getMolecules(junctionType)) / \
        network.getVolume()  # number of chains (network strands) per unit volume
    if (functionalityPerType is None or junctionType not in functionalityPerType):
        functionalityPerType = network.determineFunctionalityPerType()
    p = computeExtentOfReaction(network, junctionType, functionalityPerType)
    r = computeStoichiometricInbalance(
        network, junctionType, strandLength, functionalityPerType)
    f = functionalityPerType[junctionType]
    alpha, beta = computeMMsProbabilities(r, p, f)
    weightFractions = computeWeightFractions(network)
    return nu * k_B * T * (2*r/f) * (f - 2)/2 * weightFractions[junctionType] * alpha * (1-alpha)**f


def calculateWeightFractionOfDanglingChains(network: Universe, junctionType, strandLength: int = None, functionalityPerType: dict = None) -> float:
    """
    Compute the weight fraction of dangling strands in infinite network

    Arguments:
      - network: the network to compute the weight fraction for
      - crosslinkerType: the atom type to use to split the molecules
      - strandLength: the length of the network strands (in nr. of beads). See: #computeStoichiometricInbalance

    Returns:
      - weightFraction $\\Phi_d = 1 - \\Phi_{el}$: weightDangling/weightTotal
    """
    return 1 - calculateWeightFractionOfBackbone(network, junctionType, strandLength, functionalityPerType)


def calculateWeightFractionOfBackbone(network: Universe, junctionType, strandLength: int = None, functionalityPerType: dict = None) -> float:
    """
    Compute the weight fraction of the backbone strands in an infinite network

    Source:
      - https://pubs.acs.org/doi/suppl/10.1021/acs.macromol.0c02737 (see supporting information for formulae)

    Arguments:
      - network: the poylmer network to do the computation for
      - junctionType: the type of the junctions/crosslinkers to select them in the network
      - strandLength: the length of the network strands (in nr. of beads). See: #computeStoichiometricInbalance
      - functionalityPerType: a dictionary with key: type, and value: functionality of this atom type. 
          See: #computeExtentOfReaction

    Returns:
      - :math:`\\Phi_{el}`: weight fraction of network backbone
    """
    if (network.getNrOfAtoms() == 0):
        return 0

    if (functionalityPerType is None or junctionType not in functionalityPerType):
        functionalityPerType = network.determineFunctionalityPerType()

    W_sol, weightFractions, alpha, beta = computeWeightFractionOfSolubleMaterial(
        network, junctionType, strandLength, functionalityPerType)

    Phi_el = 0
    W_a = weightFractions[junctionType]/functionalityPerType[junctionType]
    W_xl = weightFractions[junctionType]
    W_x2 = 1-W_xl
    if (functionalityPerType[junctionType] == 3):
        Phi_el = ((W_x2*(1-beta)**2) +
                  (W_xl*((1-alpha)**3 + 3*alpha*(1-W_a)*((1-alpha)**2))))/(1-W_sol)
    else:
        assert(functionalityPerType[junctionType] == 4)
        Phi_el = ((W_x2*(1-beta)**2) +
                  (W_xl*(((1-alpha)**4) + 4*alpha*(1-W_a) * ((1-alpha)**3) +
                         6*(alpha**2)*(1-2*W_a)*(1-alpha)**2)))/(1-W_sol)

    return Phi_el


def measureWeightFractioOfSolubleMaterial(network: Universe, relTol: float = 0.75, absTol: float = None) -> float:
    """
    Compute the weight fraction of soluble material by counting.

    Arguments:
      - network: the poylmer network to do the computation for
      - relTol: the fraction of the maximum weight that counts as soluble. Ignored if absTol is specified
      - absTol: the weight from which on a component is not soluble anymore

    Returns:
      - :math:`W_{sol}` (float): the weight fraction of soluble material as counted.

    """
    if (network.getNrOfAtoms() == 0):
        return None
    fractions = network.getClusters()
    weights = np.array([f.computeWeight() for f in fractions])
    totalWeight = weights.sum()
    solubleWeight = 0
    for w in weights:
        if (absTol is not None):
            if (w < absTol):
                solubleWeight += w
        else:
            if (w < relTol*weights.max()):
                solubleWeight += w

    return solubleWeight/totalWeight


def computeWeightFractionOfSolubleMaterial(network: Universe = None, junctionType: int = 2, strandLength: int = None, functionalityPerType: dict = None, weightFractions: dict = None, r: float = None, p: float = None) -> float:
    """
    Compute the weight fraction of soluble material by MMT.

    Source:
      - https://pubs.acs.org/doi/10.1021/ma00046a021
      - https://pubs.acs.org/doi/suppl/10.1021/acs.macromol.0c02737

    Arguments:
      - network: the poylmer network to do the computation for
      - junctionType: the type of the junctions/crosslinkers to select them in the network
      - weightFractions (dict): a dictionary with key: type, and value: weight fraction of type. Pass if you want to omit the network.
      - strandLength (int): the length of the network strands (in nr. of beads). 
          See: :func:`~pylimer_tools.calc.doMMTAnalysis.computeStoichiometricInbalance`.
      - functionalityPerType (dict): a dictionary with key: type, and value: functionality of this atom type. 
          See: :func:`~pylimer_tools.calc.doMMTAnalysis.computeExtentOfReaction`.

    Returns:
      - :math:`W_{sol}` (float): the weight fraction of soluble material according to MMT.
      - weightFractions (dict): a dictionary with key: type, and value: weight fraction of type
      - :math:`\\alpha` (float): Macosko & Miller's :math:`P(F_A)`
      - :math:`\\beta` (float): Macosko & Miller's :math:`P(F_B)`
    """
    if (functionalityPerType is None or junctionType not in functionalityPerType):
        assert(network is not None)
        functionalityPerType = network.determineFunctionalityPerType()

    if (functionalityPerType[junctionType] not in [3, 4]):
        raise NotImplementedError(
            "Currently, only crosslinker functionality of 3 or 4 is supported. {} given.".format(functionalityPerType[junctionType]))

    for key in functionalityPerType:
        if (key != junctionType and functionalityPerType[key] != 2):
            raise NotImplementedError(
                "Currently, only strand functionality of 2 is supported. {} given for type {}".format(functionalityPerType[key], key))

    if (weightFractions is None):
        weightFractions = computeWeightFractions(network)

    if (p is None):
        assert(network is not None)
        p = computeExtentOfReaction(
            network, junctionType, functionalityPerType)
    if (r is None):
        assert(network is not None)
        r = computeStoichiometricInbalance(
            network, junctionType, strandLength=strandLength, functionalityPerType=functionalityPerType)

    alpha, beta = computeMMsProbabilities(
        r, p, functionalityPerType[junctionType])
    W_sol = 0
    for key in weightFractions:
        coeff = alpha if key == junctionType else beta
        W_sol += weightFractions[key] * \
            (math.pow(coeff, functionalityPerType[key]))

    return W_sol, weightFractions, alpha, beta


def computeMMsProbabilities(r, p, f):
    """
    Compute Macosko and Miller's probabilities :math:`P(F_A)` and :math:`P(F_B)`

    Sources:
      - https://pubs.acs.org/doi/10.1021/ma60050a004
      - https://doi.org/10.1021/ma60050a003

    Arguments:
      - r: the stoichiometric inbalance
      - p: the extent of reaction
      - f: the functionality of the the crosslinker

    Returns:
      - alpha: :math:`P(F_A)`
      - beta: :math:`P(F_B)`    
    """
    # first, check a few things required by the formulae
    # since we want alpha, beta \in [0,1], given they are supposed to be probabilities
    # if (r > 1 or r < 0):
    #     raise ValueError(
    #         "A stoichiometric inbalance ouside of [0, 1] is not (yet) supported. Got {}".format(r))
    # if (p < 1/math.sqrt(2) or p > 1):
    #     raise ValueError(
    #         "The extent of reaction has to be inside [1/sqrt(2), 1] for the result to be realistic. Got {}".format(p))
    # if (r <= 1/(2*p*p) and f == 3):
    #     raise ValueError(
    #         "The stoichiometric inbalance must be > 1/(2p^2) for the resulting alpha to be realisitic. Got p = {}, r = {}".format(p, r))

    # actually do the calculations
    if (f == 3):
        alpha = ((1 - r*p*p)/(r*p*p))
        beta = (r*p*alpha**2)
    elif(f == 4):
        alpha = (math.sqrt((1/(r*p*p)) - 0.75) - 0.5)
        beta = ((r*p*(alpha**3)) + 1 - r*p)
    else:
        def funToRootForAlpha(alpha):
            return r*p**2*alpha**(f-1) - alpha - r*p ** 2 + 1
        alphaSol = optimize.root_scalar(
            funToRootForAlpha, bracket=(0, 1), method='brentq')
        alpha = alphaSol.root
        beta = ((r*p*alpha**(f-1)) + 1 - r*p)  # TODO: reconsider
    return alpha, beta


def computeWeightFractions(network: Universe) -> dict:
    """
    Compute the weight fractions of each atom type in the network.

    Arguments:
      - network: the poylmer network to do the computation for

    Returns:
      - :math:`\\vec{W_i}` (dict): using the type i as a key, this dict contains the weight fractions (:math:`\\frac{W_i}{W_{tot}}`)
    """
    if (network.getNrOfAtoms() == 0):
        return {}

    weightPerType = network.getMasses()
    counts = Counter(network.getAtomTypes())
    totalMass = 0
    partialMasses = {}
    for key in counts:
        totalMass += counts[key]*weightPerType[key]
        partialMasses[key] = counts[key]*weightPerType[key]

    if (totalMass == 0):
        return partialMasses

    weightFractions = {}
    for key in partialMasses:
        weightFractions[key] = partialMasses[key]/totalMass

    return weightFractions


def computeStoichiometricInbalance(network: Universe, junctionType, strandLength: int = None, functionalityPerType: dict = None, ignoreTypes: list = [], effective: bool = False) -> float:
    """
    Compute the stoichiometric inbalance
    ( nr. of bonds formable of crosslinker / nr. of formable bonds of precursor )

    NOTE: 
      if your system has a non-integer number of possible bonds (e.g. one site unbonded),
      this will not be rounded/respected in any way. 

    Arguments:
      - network: the poylmer network to do the computation for
      - junctionType: the type of the junctions/crosslinkers to select them in the network
      - strandLength: the length of the network strands (in nr. of beads). 
          Used to infer the number of precursor strands. 
          If `None`: will use average length of each connected system when ignoring the crosslinkers.
      - functionalityPerType: a dictionary with key: type, and value: functionality of this atom type. 
          If `None`: will use max functionality per type.
      - ignoreTypes: a list of integers, the types to ignore for the inbalance (e.g. solvent atom types)
      - effective: whether to use the effective functionality (if functionalityPerType is not passed) or the maximum

    Returns:
      - r (float): the stoichiometric inbalance
    """
    if (network.getNrOfAtoms() == 0):
        return 0

    counts = Counter(network.getAtomTypes())

    if (functionalityPerType is None or junctionType not in functionalityPerType):
        functionalityPerType = network.determineEffectiveFunctionalityPerType(
        ) if effective else network.determineFunctionalityPerType()

    if (junctionType not in counts):
        raise ValueError(
            "No junction with type {} seems to have been found in the network".format(junctionType))

    if (strandLength is None):
        strands = network.getMolecules(junctionType)
        strandLength = np.mean([m.getLength() for m in strands if not np.all(
            [a.getType() in ignoreTypes for a in m.getAtoms()])])

    crosslinkerFormableBonds = counts[junctionType] * \
        functionalityPerType[junctionType]
    otherFormableBonds = 0
    for key in counts:
        if (key in ignoreTypes):
            continue
        if (key not in functionalityPerType):
            raise ValueError(
                "Type {} must have an associated functionality".format(key))
        if (key != junctionType):
            otherFormableBonds += counts[key]*functionalityPerType[key]

    # division by 2 is implicit
    return crosslinkerFormableBonds/(otherFormableBonds/strandLength)


def computeExtentOfReaction(network: Universe, crosslinkerType, functionalityPerType: dict = None, strandLength: float = None) -> float:
    """
    Compute the extent of polymerization reaction
    (nr. of formed bonds in reaction / max. nr. of bonds formable)
    NOTE: if your system has a non-integer number of possible bonds (e.g. one site unbonded),
    this will not be rounded/respected in any way. 

    Arguments:
      - network: the poylmer network to do the computation for
      - crosslinkerType: the atom type of crosslinker beads
      - functionalityPerType: a dictionary with key: type, and value: functionality of this atom type. 
          If None: will use max functionality per type.
      - strandLength: the length of the network strands (in nr. of beads). 
          If None: will compute from network structure

    Returns:
      - p (float): the extent of reaction
    """

    if (network.getNrOfAtoms() == 0):
        return 1

    if (functionalityPerType is None or crosslinkerType not in functionalityPerType):
        functionalityPerType = network.determineFunctionalityPerType()

    numStrands = len(network.getMolecules(crosslinkerType))
    numCrosslinkers = len(network.getAtomsWithType(crosslinkerType))

    # assuming strand has functionality 2
    maxFormableBonds = min(numStrands*2, numCrosslinkers *
                           functionalityPerType[crosslinkerType])

    if (maxFormableBonds == 0):
        return 1

    if (strandLength is None):
        strands = network.getMolecules(crosslinkerType)
        strandLength = np.mean([m.getLength() for m in strands])

    actuallyFormedBonds = (network.getNrOfBonds() -
                           (numStrands * (strandLength-1)))

    return actuallyFormedBonds/(maxFormableBonds)


def predictGelationPoint(r: float, f: int, g: int = 2) -> float:
    """
    Compute the gelation point :math:`p_{gel}` as theoretically predicted
    (gelation point = critical extent of reaction for gelation)

    Source:
      - https://www.sciencedirect.com/science/article/pii/003238618990253X

    Arguments:
      - r (double): the stoichiometric inbalance of reactants (see: #computeStoichiometricInbalance)
      - f (int): functionality of the crosslinkers
      - g (int): functionality of the precursor polymer

    Returns:
      - p_gel: critical extent of reaction for gelation
    """
    # if (r is None):
    #   r = calculateEffectiveCrosslinkerFunctionality(network, junctionType, f)
    return math.sqrt(1/(r*(f-1)*(g-1)))



import pandas as pd
import warnings


def unifyDataStepsizes(data: pd.DataFrame, key: str, stepSize: int = None, maxExpectedStepSize: int = 100):
    """
    Get a DataFrame where all data points have the same step between the values in column given by `key`
    NOTE: this function is rather unstable, as it has a few dirty assumptions, such as:
    - steps are modulo stepsize. Breaks e.g. with steps start with 1 and go up by stepSize.
    - ideal step-size is max step difference. Breaks e.g. if there is one big gap

    Arguments:
        - data: the DataFrame to unify the step-size for
        - key: the column name indicating the column containing the step-nr.
        - maxExpectedStepSize: use to get a warning if the computed step-size is larger

    Returns:
        - data: a DataFrame with a consistent step-size
    """
    # lenBefore = len(data)
    if (stepSize is None):
        stepSize = data[key].sort_values().diff().max()
    if (stepSize > maxExpectedStepSize):
        warnings.warn("Step size unexpectedly large")
    data = data[(data[key] % stepSize) == 0]
    # print("Reduced from {} to {} data-points using step size of {}".format(lenBefore, len(data), stepSize))
    return data



# source: https://medium.com/bigdatarepublic/advanced-pandas-optimize-speed-and-memory-a654b53be6c2,
# source: https://stackoverflow.com/questions/57531388/how-can-i-reduce-the-memory-of-a-pandas-dataframe

import gc
from typing import List

import numpy as np
import pandas as pd


def reduce_mem_usage(df, obj_to_category=False, subset=None):
    """
    Iterate through all the columns of a dataframe and modify the data type to reduce memory usage.

    Arguments:
        df (pd.DataFrame): dataframe to reduce
        obj_to_category (boolean): convert non-datetime related objects to category dtype
        subset (List): subset of columns to analyse

    Returns:
        df (pd.DataFrame): dataset with the column dtypes adjusted
    """
    start_mem = df.memory_usage().sum() / 1024 ** 2
    gc.collect()
    # print('Memory usage of dataframe is {:.2f} MB'.format(start_mem))

    cols = subset if subset is not None else df.columns.tolist()

    for col in cols:
        col_type = df[col].dtype

        if col_type != object and col_type.name != 'category' and 'datetime' not in col_type.name:
            c_min = df[col].min()
            c_max = df[col].max()

            # test if column can be converted to an integer
            treat_as_int = str(col_type)[:3] == 'int'

            if treat_as_int:
                if c_min > np.iinfo(np.int8).min and c_max < np.iinfo(np.int8).max:
                    df[col] = df[col].astype(np.int8)
                elif c_min > np.iinfo(np.uint8).min and c_max < np.iinfo(np.uint8).max:
                    df[col] = df[col].astype(np.uint8)
                elif c_min > np.iinfo(np.int16).min and c_max < np.iinfo(np.int16).max:
                    df[col] = df[col].astype(np.int16)
                elif c_min > np.iinfo(np.uint16).min and c_max < np.iinfo(np.uint16).max:
                    df[col] = df[col].astype(np.uint16)
                elif c_min > np.iinfo(np.int32).min and c_max < np.iinfo(np.int32).max:
                    df[col] = df[col].astype(np.int32)
                elif c_min > np.iinfo(np.uint32).min and c_max < np.iinfo(np.uint32).max:
                    df[col] = df[col].astype(np.uint32)
                elif c_min > np.iinfo(np.int64).min and c_max < np.iinfo(np.int64).max:
                    df[col] = df[col].astype(np.int64)
                elif c_min > np.iinfo(np.uint64).min and c_max < np.iinfo(np.uint64).max:
                    df[col] = df[col].astype(np.uint64)
            else:
                if c_min > np.finfo(np.float16).min and c_max < np.finfo(np.float16).max:
                    df[col] = df[col].astype(np.float16)
                elif c_min > np.finfo(np.float32).min and c_max < np.finfo(np.float32).max:
                    df[col] = df[col].astype(np.float32)
                else:
                    df[col] = df[col].astype(np.float64)
        elif 'datetime' not in col_type.name and obj_to_category:
            df[col] = df[col].astype('category')
    gc.collect()
    end_mem = df.memory_usage().sum() / 1024 ** 2
    # print('Memory usage after optimization is: {:.3f} MB'.format(end_mem))
    # print('Decreased by {:.1f}%'.format(
    #     100 * (start_mem - end_mem) / start_mem))

    return df


def optimize_floats(df: pd.DataFrame) -> pd.DataFrame:
    """
    Optimize the floating point type entries

    Arguments:
        df (pd.DataFrame): dataframe to reduce

    Returns:
        df (pd.DataFrame): dataset with the column dtypes adjusted
    """
    floats = df.select_dtypes(include=['float64']).columns.tolist()
    df[floats] = df[floats].apply(pd.to_numeric, downcast='float')
    return df


def optimize_ints(df: pd.DataFrame) -> pd.DataFrame:
    """
    Optimize the integer point type entries

    Arguments:
        df (pd.DataFrame): dataframe to reduce

    Returns:
        df (pd.DataFrame): dataset with the column dtypes adjusted
    """
    ints = df.select_dtypes(include=['int64']).columns.tolist()
    df[ints] = df[ints].apply(pd.to_numeric, downcast='integer')
    return df


def optimize_objects(df: pd.DataFrame, datetime_features: List[str]) -> pd.DataFrame:
    """
    Optimize object type entries

    Arguments:
        - df (pd.DataFrame): dataframe to reduce

    Returns:
        - df (pd.DataFrame): dataset with the column dtypes adjusted
    """
    for col in df.select_dtypes(include=['object']):
        if col not in datetime_features:
            num_unique_values = len(df[col].unique())
            num_total_values = len(df[col])
            if float(num_unique_values) / num_total_values < 0.5:
                df[col] = df[col].astype('category')
        else:
            df[col] = pd.to_datetime(df[col])
    return df


def optimize(df: pd.DataFrame, datetime_features: List[str] = []):
    """
    Optimize all types of all columns in a dataframe

    Arguments:
        - df (pd.DataFrame): dataframe to reduce

    Returns:
        - df (pd.DataFrame): dataset with the column dtypes adjusted
    """
    return optimize_floats(optimize_ints(optimize_objects(df, datetime_features)))




import csv
from datetime import datetime
import hashlib
import os
import pathlib
import pickle
from io import StringIO
from typing import Iterable
import warnings
import tempfile

import pandas as pd
from pylimer_tools.utils.cacheUtility import doCache, loadCache
from pylimer_tools.utils.optimizeDf import optimize, reduce_mem_usage


# Helper functions
def readOneGroup(fp, header, minLineLen=4, additional_lines_skip=0) -> str:
    """
    Read one group of csv lines from the file

    Arguments:
        - fp: the file pointer to the file to read from
        - header: the header of the CSV (where to start reading at)
        - minLineLen: the minimal length of a line to be accepted as data
        - additional_lines_skip: number of lines to skip after reading the header


    Returns:
       The filename of a temporary CSV file
    """
    csvFileToWrite = "{}/{}_{}".format(
        tempfile.gettempdir(),
        hashlib.md5(datetime.now().strftime("%m.%d.%Y, %H:%M:%S").encode()).hexdigest(), 'tmp_thermo_file.csv')
    n_lines = 0
    with open(csvFileToWrite, 'w') as output_csv:
        line = fp.readline()
        separator = ", "
        headerLen = None
        if (isinstance(header, str)):
            minLineLen = max(minLineLen, len(header.split()))
        else:
            minLineLen = max(minLineLen, min([len(h.split()) for h in header]))

        def checkSkipLine(line, header):
            return line and not line.startswith(header)

        def checkSkipLineHeaderList(line, header):
            if (not line):
                return False
            for headerL in header:
                if (line.startswith(headerL)):
                    return False
            return True

        skipLineFun = checkSkipLineHeaderList if isinstance(
            header, list) else checkSkipLine
        # skip lines up until header (or file ending)
        while skipLineFun(line, header):
            line = fp.readline()
        # found header. Take next few lines:
        headerLen = len(line.split())
        if (not line):
            return ""
        else:
            output_csv.write((separator.join(line.split())).strip() + "\n")

        n_lines = 0
        while line and n_lines < additional_lines_skip:
            # skip ${additional_lines_skip} further
            line = fp.readline()
            # text += (', '.join(line.split())).strip() + "\n"
            n_lines += 1
        while line and not line.startswith("Loop time of"):
            line = fp.readline()
            if (len(line) < minLineLen or (len(line.split()) != headerLen) or (len(line) > 0 and (
                    line.startswith("WARNING") or
                    line[0].isalpha() or
                    (line[0] == "-" and line[1] == "-") or
                    (line[2].isalpha() or line[3].isalpha()) or
                    (line[0] == "[") or
                    ("src" in line) or
                    ("fene" in line or ")" in line)  # from ":90)"
            ))):
                # skip line due to error, warning or similar
                continue
            output_csv.write((separator.join(line.split())).strip() + "\n")
            n_lines += 1
    return csvFileToWrite if n_lines > 0 else ""


def getThermoCacheNameSuffix(header="Step Temp E_pair E_mol TotEng Press", textsToRead=5, minLineLen=5) -> str:
    """
    Compose a cache file suffix in such a way, that it distinguishes different thermo reader parameters

    Arguments:
        - header: the header of the CSV (where to start reading at)
        - textsToRead: the number of times to expect the header
        - minLineLen: the minimal length of a line to be accepted as data
    """
    if (isinstance(header, Iterable)):
        header = "{}{}".format("".join(header), len(header))

    return "{}{}{}-thermo-param-cache.pickle".format(header, textsToRead, minLineLen)


def extractThermoParams(file, header="Step Temp E_pair E_mol TotEng Press", textsToRead=5, minLineLen=5, useCache=True) -> pd.DataFrame:
    """
    Extract the thermodynamic outputs produced for this simulation.

    Note: the header parameter can be an array — make sure to pay attention
    when reading a file with different header sections in them

    Arguments:
        - file: the file path to the file to read from
        - header: the header of the CSV (where to start reading at)
        - textsToRead: the number of times to expect the header
        - minLineLen: the minimal length of a line to be accepted as data
        - useCache: wheter to use cache or not (though it will be written anyway)

    Returns:
        - data (pd.DataFrame): the thermodynamic parameters

    """
    df = None

    suffix = getThermoCacheNameSuffix(
        header, textsToRead, minLineLen)
    cacheContent = loadCache(file, suffix)

    if (cacheContent is not None and useCache):
        return cacheContent

    def csvFileToDf(filePath) -> pd.DataFrame:
        try:
            tmpDf = pd.read_csv(filePath, low_memory=False,
                               on_bad_lines='skip', quoting=csv.QUOTE_NONE)
            try:
                os.remove(filePath)
            except Exception as e:
                pass
            return tmpDf
        except Exception as e:
            warnings.warn("Error reading CSV thermo file: {}".format(e), source=e)
            return pd.DataFrame()

    with open(file, 'r') as fp:
        tmpCsvFile = readOneGroup(fp, header, minLineLen=minLineLen)
        textsRead = 1
        df = csvFileToDf(tmpCsvFile)
        while(textsRead < textsToRead):
            tmpCsvFile = readOneGroup(fp, header, minLineLen=minLineLen)
            textsRead += 1
            if (tmpCsvFile != ""):
                newDf = csvFileToDf(tmpCsvFile)
                if (not newDf.empty):
                    df = df.append(newDf)

    if (df is not None):
        # df.columns = df.columns.str.replace(' ', '')
        df.rename(columns=lambda x: x.strip(), inplace=True)
    else:
        df = pd.DataFrame()

    doCache(df, file, suffix)
    # print("Read {} rows for file {}".format(len(df), file))

    return df



import datetime
import hashlib
import os
import pathlib
import pickle
import tempfile
import warnings


def doCache(obj, file: str, suffix: str):
    """
    Store the object in the cache

    Arguments:
        - obj: the object to cache
        - file: the path to the file to save the object to
        - suffix: the file name's suffix
    """
    cacheFileName = getCacheFileName(file, suffix)
    with open(cacheFileName, 'wb') as cacheFile:
        pickle.dump(obj, cacheFile)


def loadCache(file: str, suffix: str, disableWarnings: bool = False):
    """
    Load an object from cache.

    Arguments:
        - file: a cache name. Ideally the file that is read, such that the filemtime of `file` can be used to check whether cache must be generated anew
        - suffix: the file name's suffix
        - disableWarnings: whether to disable warnings about missing possibilities to check for filemtime

    Returns:
        - cache: either the content of the cache, or None if the cache has to be loaded again / is non existant
    """
    cacheFileName = getCacheFileName(file, suffix)
    if (os.path.isfile(cacheFileName)):
        if (not os.path.isfile(file)):
            if (not disableWarnings):
                warnings.warn(
                    'Cache called for non-existent file. Make sure the key is time-restricted')
            with open(cacheFileName, 'rb') as cacheFile:
                toReturn = pickle.load(cacheFile)
            return toReturn
        else:
            mtimeCache = datetime.datetime.fromtimestamp(
                pathlib.Path(cacheFileName).stat().st_mtime)
            mtimeOrigin = datetime.datetime.fromtimestamp(
                pathlib.Path(file).stat().st_mtime)
            if (mtimeCache > mtimeOrigin):
                with open(cacheFileName, 'rb') as cacheFile:
                    toReturn = pickle.load(cacheFile)
                return toReturn
            else:
                # print("Dump cache file is elder than dump. Reloading...")
                pass

    return None


def getCacheFileName(file: str, suffix: str):
    """
    Get the name and path of a cache file. Internal method.

    Arguments:
        - file: a cache name. Ideally the file that is read.
        - suffix: the file name's suffix

    Returns:
        - cacheFileName: the path to the cache file
    """
    cacheFileName = "{}/{}-{}.pickle".format(
        tempfile.gettempdir(),
        hashlib.md5(file.encode()).hexdigest(), suffix)
    return cacheFileName



"""
This is the C++ implementation of pylimer_tools. 

The subpackage (actually compiled code) :module:`pylimer_tools_cpp.pylimer_tools_cpp` 
should be imported automatically when you `import pylimer_tools_cpp`, 
meaning there is no need to use the double `pylimer_tools_cpp.pylimer_tools_cpp`,
a single one is sufficient.
"""

from .pylimer_tools_cpp import *



#!/usr/bin/env python
# cli.py
import click
import numpy as np

from pylimer_tools_cpp import UniverseSequence


@click.command()
@click.argument('files', nargs=-1, type=click.Path(exists=True))
def cli(files):
    """
    Basic CLI application reading all passed files, outputting some stats on the structures therein

    Arguments:
      - files: list of files to read
    """
    click.echo("Processing {} files".format(len(files)))
    for filePath in files:
        click.echo("\nAnalysing File " + filePath)

        universeSequence = UniverseSequence()
        universeSequence.initializeFromDataSequence([filePath])
        universe = universeSequence.atIndex(0)
        click.echo("Size: {}. Volume: {} u^3".format(
            universe.getNrOfAtoms(), universe.getVolume()))
        molecules = universe.getMolecules(2)
        bondLengths = [np.mean(m.computeBondLengths()) for m in molecules]
        nonNoneBondLengths = [
            l for l in bondLengths if l is not None and l > 0]
        click.echo("Mean bond length: {} u, (min: {}, max: {}, median: {}) u".format(
            np.mean(nonNoneBondLengths), np.min(nonNoneBondLengths), np.max(nonNoneBondLengths), np.median(nonNoneBondLengths)))
        endToEndDistances = [m.computeEndToEndDistance() for m in molecules]
        click.echo("Mean end to end distance: {} u".format(
            np.mean([e for e in endToEndDistances if e is not None and e > 0])))
        click.echo("For {} molecules of mean length of {} atoms".format(
            len(molecules), np.mean([m.getNrOfAtoms() for m in molecules])))
    click.echo("Arbitrary units used. E.g.: Length: u")


if __name__ == "__main__":
    cli()



import pandas as pd


def getTail(data, percentage=0.2, minN=25, maxPercentage=0.5):
    """
    Extract the last few entries of a list

    Arguments:
        - data (list|pd.DataFrame|pd.Series): the list to extract the last few entries from
        - percentage: the percentage of entries to extract
        - minN: the minimum number of entries to extract
        - maxPercentage: the maximum percentage of entries to extract

    Returns:
        A (list|pd.DataFrame|pd.Series) with at maximum maxPercentage, 
            at least minN entries (assuming the initial data is as large), 
            but ideally `percentage` many percentage of the last entries.
    """
    assert(percentage <= 1)
    assert(maxPercentage <= 1)
    tailN = int(min(max(min(minN, maxPercentage*len(data)),
                        percentage*len(data)), len(data)))
    if (isinstance(data, pd.DataFrame) or isinstance(data, pd.Series)):
        return data.tail(tailN)
    else:
        return data[-tailN:]
